"""Handle Pydantic BaseModel type."""

from collections.abc import Mapping
from typing import Any

from pydantic import BaseModel
from pydantic.fields import FieldInfo

from fastlife.adapters.jinjax.widget_factory.base import BaseWidgetBuilder
from fastlife.adapters.jinjax.widgets.base import Widget
from fastlife.adapters.jinjax.widgets.model import ModelWidget


class ModelBuilder(BaseWidgetBuilder[Mapping[str, Any]]):
    """Builder for Pydantic BaseModel values."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for Pydantic BaseModel."""
        return issubclass(typ, BaseModel)

    def build(
        self,
        *,
        field_name: str,
        field_type: type[BaseModel],
        field: FieldInfo | None,
        value: Mapping[str, Any] | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> Widget[Any]:
        """Build the widget."""
        value = value or {}
        ret: dict[str, Any] = {}
        for key, child_field in field_type.model_fields.items():
            child_key = f"{field_name}.{key}" if field_name else key
            if child_field.exclude:
                continue
            if child_field.annotation is None:
                raise ValueError(  # coverage: ignore
                    f"Missing annotation for {child_field} in {child_key}"
                )
            ret[key] = self.factory.build(
                child_field.annotation,
                name=child_key,
                field=child_field,
                value=value.get(key),
                form_errors=form_errors,
                removable=False,
            )
        return ModelWidget[Any](
            name=field_name,
            value=list(ret.values()),
            removable=removable,
            title=field.title or "" if field and field.title else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            error=form_errors.get(field_name),
            nested=field is not None,
        )



"""Settings for the fastlife."""

from datetime import timedelta
from typing import Literal

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """
    Settings based on pydantic settings to configure the app.

    It can be overriden in order to inject application own settings.
    """

    model_config = SettingsConfigDict(env_prefix="fastlife_")
    """
    Set the prefix ``fastlife_`` for configuration using operating system environment.
    """

    fastlife_route_prefix: str = Field(default="/_fl")
    """Route prefix used for fastlife internal views."""
    template_search_path: str = Field(default="fastlife:components")
    """
    list of directories where components could be found by the template engine.

    the list is a comma separated string. The directory resolution is made from
    a python module name. for instance `fastlife:components` is the directory components
    found in the fastlife package.
    """
    registry_class: str = Field(default="fastlife.config.registry:DefaultRegistry")
    """Implementation class for the application regitry."""
    template_renderer_class: str = Field(
        default="fastlife.templates.renderer:JinjaxEngine"
    )
    """
    Implementation class for the
    {class}`fastlife.templates.renderer.AbstractTemplateRenderer`.
    """
    form_data_model_prefix: str = Field(default="payload")
    """
    Pydantic form default model prefix for serialized field in www-urlencoded-form.
    """
    csrf_token_name: str = Field(default="csrf_token")
    """
    Name of the html input field and for the http cookie for csrf token.
    """

    default_locale: str = Field(default="en")
    """
    The default locale
    """

    jinjax_file_ext: str = Field(default="jinja")
    """
    JinjaX file extention.
    """

    jinjax_use_cache: bool = Field(default=True)
    """
    JinjaX (the default template engine) setting use_cache.

    Could be disabled while developping, leave value true for production.
    """
    jinjax_auto_reload: bool = Field(default=False)
    """
    JinjaX (the default template engine) setting auto_reload.

    Set to true while developing, set false for production.
    """
    jinjax_global_catalog_class: str = Field(
        default="fastlife.templates.constants:Constants"
    )
    """
    Set global constants accessible in every templates.
    Defaults to `fastlife.templates.constants:Constants`
    See {class}`fastlife.templates.constants.Constants`
    """

    session_secret_key: str = Field(default="")
    """
    A secret key, that could not changes for life.

    Used to securize the session with itsdangerous.
    """
    session_cookie_name: str = Field(default="flsess")
    """Cookie name for the session cookie."""
    session_cookie_domain: str = Field(default="")
    """Cookie domain for the session cookie."""
    session_cookie_path: str = Field(default="/")
    """Cookie path for the session cookie."""
    session_duration: timedelta = Field(default=timedelta(days=14))
    """Cookie duration for the session cookie."""
    session_cookie_same_site: Literal["lax", "strict", "none"] = Field(default="lax")
    """Cookie same-site for the session cookie."""
    session_cookie_secure: bool = Field(default=False)
    """
    Cookie secure for the session cookie,

    should be true while using https on production.
    """
    session_serializer: str = Field(
        default="fastlife.middlewares.session.serializer:SignedSessionSerializer"
    )
    """Cookie serializer for the session cookie."""

    domain_name: str = Field(default="", title="domain name where the app is served")
    """Domain name whre the app is served."""

    decode_reverse_proxy_headers: bool = Field(default=True)
    """Ensure that the request object has information based on http proxy headers."""



"""Inline templates."""

from typing import ClassVar

from pydantic import BaseModel
from pydantic.config import ConfigDict


class InlineTemplate(BaseModel):
    """
    Inline templates are used to encourage the location of behavior and the view typing.

    Pages produce templates that are not reusable and don't need to be reusable
    in there essence, they don't need to be in a component library.
    They use a component lirary to stay small but contains a view logic
    tighly coupled with the view and its code can stay in the same module of that view.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    template: ClassVar[str]
    """The template string to render."""
    renderer: ClassVar[str]
    """Template render engine to use."""


class JinjaXTemplate(InlineTemplate):
    """Template that render JinjaX"""

    renderer = ".jinja"



"""
Base class to of the template renderer.

Fastlife comes with {class}`fastlife.adapters.jinjax.renderer.JinjaxEngine`,
the rendering engine.

More template engine can be registered using the configurator method
{meth}`add_renderer <fastlife.config.configurator.GenericConfigurator.add_renderer>`
"""

import abc
from collections.abc import Mapping
from typing import Any

from fastlife import Request, Response
from fastlife.templates.inline import InlineTemplate

TemplateParams = Mapping[str, Any]


class AbstractTemplateRenderer(abc.ABC):
    """
    An object that will be initialized by an AbstractTemplateRendererFactory,
    passing the request to process.
    """

    request: Request
    """Associated request that needs a response."""

    def __init__(self, request: Request) -> None:
        self.request = request
        self.globals: dict[str, Any] = {}

    @property
    def route_prefix(self) -> str:
        """Used to buid pydantic form widget that do ajax requests."""
        return self.request.registry.settings.fastlife_route_prefix

    def render(
        self,
        template: str,
        *,
        status_code: int = 200,
        content_type: str = "text/html",
        globals: Mapping[str, Any] | None = None,
        params: InlineTemplate,
    ) -> Response:
        """
        Render the template and build the HTTP Response.
        """
        request = self.request
        if globals:
            self.globals.update(globals)
        data = self.render_template(params)
        resp = Response(
            data, status_code=status_code, headers={"Content-Type": content_type}
        )
        resp.set_cookie(
            request.csrf_token.name,
            request.csrf_token.value,
            secure=request.url.scheme == "https",
            samesite="strict",
            max_age=60 * 15,
        )
        return resp

    @abc.abstractmethod
    def render_template(self, template: InlineTemplate) -> str:
        """
        Render an inline template.

        :param template: the template to render.
        :return: The template rendering result.
        """


class AbstractTemplateRendererFactory(abc.ABC):
    """
    The template render factory.

    The implementation of this class is found using the settings
    :attr:`fastlife.config.settings.Settings.template_renderer_class`.
    """

    @abc.abstractmethod
    def __call__(self, request: Request) -> AbstractTemplateRenderer:
        """
        While processing an HTTP Request, a renderer object is created giving
        isolated context per request.

        :param Request: the HTTP Request to process.
        :return: The renderer object that will process that request.
        """



"""Inline templates."""

import inspect
from collections.abc import Callable
from typing import Any, get_args

from fastlife.domain.model.template import InlineTemplate
from fastlife.shared_utils.infer import is_union


def is_inline_template_returned(endpoint: Callable[..., Any]) -> bool:
    signature = inspect.signature(endpoint)
    return_annotation = signature.return_annotation

    if isinstance(return_annotation, type) and issubclass(
        return_annotation, InlineTemplate
    ):
        return True

    if is_union(return_annotation):
        return any(
            isinstance(arg, type) and issubclass(arg, InlineTemplate)
            for arg in get_args(return_annotation)
        )

    return False



"""
Widget for field of type Enum or Literal.
"""

from collections.abc import Sequence

from pydantic import Field, field_validator

from .base import Widget


class DropDownWidget(Widget[str]):
    """
    Widget for field of type Enum or Literal.
    """

    template = """
    <pydantic_form.Widget :widget_id="id" :removable="removable">
      <div class="pt-4">
        <Label :for="id">{{title}}</Label>
        <Select :name="name" :id="id">
          {%- for opt in options -%}
          <Option :value="opt.value" id={{id + "-" + opt.value.replace(" ", " -")}}
            :selected="value==opt.value">
            {{- opt.text -}}
          </Option>
          {%- endfor -%}
        </Select>
        <pydantic_form.Error :text="error" />
        <pydantic_form.Hint :text="hint" />
      </div>
    </pydantic_form.Widget>
    """

    options: list[dict[str | int, str | int]] = Field(default_factory=list)

    @field_validator("options", mode="before")
    @classmethod
    def validate_options(
        cls, options: Sequence[str | int | tuple[str | int, str | int]] | None
    ) -> Sequence[dict[str | int, str | int]]:
        if not options:
            return []
        ret: list[dict[str | int, str | int]] = []
        for opt in options:
            if isinstance(opt, tuple):
                ret.append({"value": opt[0], "text": opt[1]})
            else:
                ret.append({"value": opt, "text": opt})
        return ret



"""Types for OpenAPI documentation."""

from pydantic import BaseModel, Field


class ExternalDocs(BaseModel):
    """OpenAPI externalDocs object."""

    description: str
    """link's description."""
    url: str
    """link's URL."""


class OpenApiTag(BaseModel):
    """OpenAPI tag object."""

    name: str
    """name of the tag."""
    description: str
    """explanation of the tag."""
    external_docs: ExternalDocs | None = Field(alias="externalDocs", default=None)
    """external link to the doc."""



"""
Template rending based on JinjaX.
"""

import logging
import re
from collections.abc import Iterator, Sequence

from jinja2.exceptions import TemplateSyntaxError
from jinjax import DuplicateDefDeclaration, InvalidArgument
from jinjax.catalog import Catalog

from .inspectable_component import InspectableComponent

log = logging.getLogger(__name__)


def to_include(
    name: str,
    ignores: Sequence[re.Pattern[str]] | None = None,
    includes: Sequence[re.Pattern[str]] | None = None,
) -> bool:
    if includes and not any(include.match(name) for include in includes):
        return False

    if ignores and any(ignore.match(name) for ignore in ignores):
        return False

    return True


class InspectableCatalog(Catalog):
    """
    JinjaX Catalog with introspection support.

    Override the catalog in order to iterate over components to build the doc.
    """

    def iter_components(
        self,
        ignores: Sequence[re.Pattern[str]] | None = None,
        includes: Sequence[re.Pattern[str]] | None = None,
    ) -> Iterator[InspectableComponent]:
        """
        Walk into every templates from the settings, iterate over components.

        :params ignores: filter components using an block list.
        :params includes: filter components using an allowed list.
        """
        for prefix, loader in self.prefixes.items():
            for t in loader.list_templates():
                name, file_ext = t.split(".", maxsplit=1)
                name = name.replace("/", ".")
                path, tmpl_name = self._get_component_path(
                    prefix, name, file_ext=file_ext
                )
                is_included = to_include(name, ignores, includes)
                if is_included:
                    try:
                        component = InspectableComponent(
                            name=name, prefix=prefix, path=path, source=path.read_text()
                        )
                    except InvalidArgument as exc:
                        log.error(f"Definition Syntax Error in <{name}/>: {exc}")
                        log.error(path.read_text())
                        continue
                    except DuplicateDefDeclaration as exc:
                        log.error(f"Duplicate Definition in <{name}/>: {exc}")
                        log.error(path.read_text())
                        continue

                    self.jinja_env.loader = loader

                    try:
                        component.tmpl = self.jinja_env.get_template(
                            tmpl_name, globals=self._tmpl_globals
                        )
                    except TemplateSyntaxError as exc:
                        log.error(
                            f"Template Syntax Error in <{name}/>: {exc} on "
                            "{exc.lineno} :"
                        )
                        log.error(path.read_text())
                        continue
                    yield component



import logging
from collections.abc import Sequence

from starlette.types import ASGIApp, Receive, Scope, Send

from fastlife.middlewares.base import AbstractMiddleware

log = logging.getLogger(__name__)


def get_header(headers: Sequence[tuple[bytes, bytes]], key: bytes) -> str | None:
    for hkey, val in headers:
        if hkey.lower() == key:
            return val.decode("latin1")
    return None


def get_header_int(headers: Sequence[tuple[bytes, bytes]], key: bytes) -> int | None:
    for hkey, val in headers:
        if hkey.lower() == key:
            try:
                return int(val.decode("latin1"))
            except ValueError:
                break
    return None


class XForwardedStar(AbstractMiddleware):
    def __init__(
        self,
        app: ASGIApp,
    ) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] in ("http", "websocket"):
            headers = scope["headers"]
            x_real_ip = get_header(headers, b"x-real-ip")
            client = (
                (
                    x_real_ip,
                    get_header_int(headers, b"x-real-port")
                    or get_header_int(headers, b"x-forwarded-port")
                    or 0,
                )
                if x_real_ip
                else None
            )
            new_vals = {
                "client": client,
                "host": get_header(headers, b"x-forwarded-host"),
                "scheme": get_header(headers, b"x-forwarded-proto"),
            }
            scope.update({key: val for key, val in new_vals.items() if val is not None})

        await self.app(scope, receive, send)



"""
FastApi router for fastlife application.

The aim of this router is get {class}`fastlife.routing.route.Route`
available in the FastApi request depency injection.
"""

from typing import Any

from fastapi import APIRouter

from fastlife.routing.route import Route


class Router(APIRouter):
    """
    The router used split your app in many routes.
    """

    def __init__(self, **kwargs: Any) -> None:
        kwargs["route_class"] = Route
        super().__init__(**kwargs)



"""Configure fastlife app for dependency injection."""

from .configurator import Configurator, GenericConfigurator, configure
from .registry import DefaultRegistry, GenericRegistry
from .resources import resource, resource_view
from .settings import Settings
from .views import view_config

__all__ = [
    "Configurator",
    "GenericConfigurator",
    "configure",
    "view_config",
    "resource",
    "resource_view",
    "GenericRegistry",
    "DefaultRegistry",
    "Settings",
]



"""HTTP Request representation in a python object."""

from typing import TYPE_CHECKING, Annotated, Any, Generic

from fastapi import Request as FastAPIRequest
from fastapi.params import Depends

from fastlife.config.registry import DefaultRegistry, TRegistry
from fastlife.domain.model.security import CSRFToken, create_csrf_token

if TYPE_CHECKING:
    from fastlife.security.policy import (  # coverage: ignore
        AbstractSecurityPolicy,
        HasPermission,
    )


class GenericRequest(FastAPIRequest, Generic[TRegistry]):
    """HTTP Request representation."""

    registry: TRegistry
    """Direct access to the application registry."""
    locale_name: str
    """Request locale used for the i18n of the response."""

    security_policy: "AbstractSecurityPolicy[Any, TRegistry] | None"
    """Request locale used for the i18n of the response."""

    renderer_globals: dict[str, Any]

    def __init__(self, registry: TRegistry, request: FastAPIRequest) -> None:
        super().__init__(request.scope, request.receive)
        self.registry = registry
        self.locale_name = registry.locale_negociator(self)
        self.security_policy = None  # build it from the ? registry
        self.renderer_globals = {}
        self._csrf_token: CSRFToken | None = None

    @property
    def csrf_token(self) -> CSRFToken:
        if self._csrf_token is None:
            name = self.registry.settings.csrf_token_name
            value = self.cookies.get(name) or create_csrf_token()
            self._csrf_token = CSRFToken(name=name, value=value)
        return self._csrf_token

    def add_renderer_globals(self, **kwargs: Any) -> None:
        """
        Add global variables to the template renderer context for the current request.
        """
        self.renderer_globals.update(kwargs)

    async def has_permission(
        self, permission: str
    ) -> "HasPermission | type[HasPermission]":
        """
        A helper to check that a user has the given permission.

        Not that this method does not raised, it return a boolean like object.
        It allows batch permission checks.
        You might need to check multiple permissions in different contexts or
        for different resources before raising an http error.
        """
        if self.security_policy is None:
            raise RuntimeError(
                f"Request {self.url.path} require a security policy, "
                "explicit fastlife.security.policy.InsecurePolicy is required."
            )

        return await self.security_policy.has_permission(permission)


def get_request(request: FastAPIRequest) -> GenericRequest[Any]:
    return request  # type: ignore


Request = Annotated[GenericRequest[DefaultRegistry], Depends(get_request)]
"""A request that is associated to the default registry."""
# FastAPI handle its Request objects using a lenient_issubclass,
# basically a issubclass(Request), doe to the Generic[T], it does not work.


AnyRequest = Annotated[GenericRequest[Any], Depends(get_request)]
"""A request version that is associated to the any registry."""


def get_registry(request: Request) -> DefaultRegistry:
    return request.registry


Registry = Annotated[DefaultRegistry, Depends(get_registry)]
"""FastAPI dependency to access to the registry."""



from typing import cast

from fastapi import Query
from pydantic.fields import FieldInfo

from fastlife import Configurator, Request, Response, configure
from fastlife.adapters.jinjax.renderer import JinjaxRenderer
from fastlife.shared_utils.resolver import resolve_extended


async def show_widget(
    typ: str,
    request: Request,
    title: str | None = Query(None),
    name: str | None = Query(None),
    token: str | None = Query(None),
    removable: bool = Query(False),
) -> Response:
    """
    This views is used by pydantic_form to generate a nested field asynchronously.
    """
    model_cls = resolve_extended(typ)
    field = None
    if title:
        field = FieldInfo(title=title)
    # FIXME: .jinja should not be hardcoded
    renderer = cast(JinjaxRenderer, request.registry.get_renderer(".jinja")(request))
    data = renderer.pydantic_form_field(
        model=model_cls,  # type: ignore
        name=name,
        token=token,
        removable=removable,
        field=field,
    )
    return Response(data, headers={"Content-Type": "text/html"})


@configure
def includeme(config: Configurator) -> None:
    route_prefix = config.registry.settings.fastlife_route_prefix
    config.add_route(
        "fl-pydantic-form-widget",
        f"{route_prefix}/pydantic-form/widgets/{{typ}}",
        show_widget,
        methods=["GET"],
    )



"""HTTP Form serialization."""

from collections.abc import Callable, Mapping
from typing import Any, Generic, TypeVar, get_origin

from fastapi import Depends
from pydantic import BaseModel, ValidationError

from fastlife import Registry
from fastlife.request.form_data import MappingFormData
from fastlife.shared_utils.infer import is_union

T = TypeVar("T", bound=BaseModel)
"""Template type for form serialized model"""


class FormModel(Generic[T]):
    prefix: str
    model: T
    errors: Mapping[str, str]
    is_valid: bool

    def __init__(
        self, prefix: str, model: T, errors: Mapping[str, Any], is_valid: bool = False
    ) -> None:
        self.prefix = prefix
        self.model = model
        self.errors = errors
        self.is_valid = is_valid

    @classmethod
    def default(cls, prefix: str, pydantic_type: type[T]) -> "FormModel[T]":
        return cls(prefix, pydantic_type.model_construct(), {})

    def edit(self, pydantic_type: T) -> None:
        """
        Load the form with the given model and consider it as valid for the user.

        No error will be reported.
        """
        self.model = pydantic_type
        self.errors = {}
        self.is_valid = True

    @property
    def form_data(self) -> Mapping[str, Any]:
        return {self.prefix: self.model.model_dump()}

    @classmethod
    def from_payload(
        cls, prefix: str, pydantic_type: type[T], data: Mapping[str, Any]
    ) -> "FormModel[T]":
        try:
            return cls(prefix, pydantic_type(**data.get(prefix, {})), {}, True)
        except ValidationError as exc:
            errors: dict[str, str] = {}
            for error in exc.errors():
                loc = prefix
                typ: Any = pydantic_type
                for part in error["loc"]:
                    if isinstance(part, str):
                        type_origin = get_origin(typ)
                        if type_origin:
                            if is_union(typ):
                                args = typ.__args__
                                for arg in args:
                                    if arg.__name__ == part:
                                        typ = arg
                                        continue

                            else:
                                raise NotImplementedError from exc  # coverage: ignore
                        elif issubclass(typ, BaseModel):
                            typ = typ.model_fields[part].annotation
                            loc = f"{loc}.{part}"
                        else:
                            raise NotImplementedError from exc  # coverage: ignore

                    else:
                        # it is an integer and it part of the list
                        loc = f"{loc}.{part}"

                if loc in errors:
                    errors[loc] = f"{errors[loc]}, {error['msg']}"
                else:
                    errors[loc] = error["msg"]
            model = pydantic_type.model_construct(**data.get(prefix, {}))
            return cls(prefix, model, errors)


def form_model(
    cls: type[T], name: str | None = None
) -> Callable[[Mapping[str, Any]], FormModel[T]]:
    """
    Build a model, a class of type T based on Pydandic Base Model from a form payload.
    """

    def to_model(data: MappingFormData, registry: Registry) -> FormModel[T]:
        prefix = name or registry.settings.form_data_model_prefix
        if not data:
            return FormModel[T].default(prefix, cls)
        return FormModel[T].from_payload(prefix, cls, data)

    return Depends(to_model)



from importlib import metadata

__version__ = metadata.version("fastlifeweb")

from fastapi import Response
from fastapi.responses import RedirectResponse

from .config import (
    Configurator,
    DefaultRegistry,
    GenericConfigurator,
    GenericRegistry,
    Settings,
    configure,
    resource,
    resource_view,
    view_config,
)
from .domain.model.template import JinjaXTemplate
from .request import GenericRequest, Registry, Request, get_request

# from .request.form_data import model
from .services.templates import TemplateParams

__all__ = [
    # Config
    "configure",
    "GenericConfigurator",
    "Configurator",
    "DefaultRegistry",
    "GenericRegistry",
    "TemplateParams",
    "Settings",
    "view_config",
    "resource",
    "resource_view",
    # Model
    # "model",
    "Request",
    "GenericRequest",
    "get_request",
    "Registry",
    "Response",
    "RedirectResponse",
    # Template
    "JinjaXTemplate",
]



"""Class utilities to access to the web form."""

from typing import TYPE_CHECKING, Any

from multidict import MultiDict

if TYPE_CHECKING:
    from .dom import Element  # coverage: ignore
    from .testclient import WebResponse, WebTestClient  # coverage: ignore


class WebForm:
    """
    Handle html form.

    Form are filled out and submit with methods and try to avoid invalid
    usage, such as selecting an option that don't exists is not possible here.
    Again, no javascript is executed here, but htmx attribute `hx-post` and `hx-target`
    are read while submiting to simulate it.
    """

    def __init__(self, client: "WebTestClient", origin: str, form: "Element"):
        self._client = client
        self._form = form
        self._origin = origin
        self._formfields: dict[str, Element] = {}
        self._formdata: MultiDict[str] = MultiDict()
        inputs = self._form.by_node_name("input")
        for input in inputs:
            self._formfields[input.attrs["name"]] = input
            if input.attrs.get("type") == "checkbox" and "checked" not in input.attrs:
                continue
            self._formdata.add(input.attrs["name"], input.attrs.get("value", ""))

        selects = self._form.by_node_name("select")
        for select in selects:
            fieldname = select.attrs["name"]
            self._formfields[fieldname] = select
            options = select.by_node_name("option")
            if "multiple" in select.attrs:
                for option in options:
                    if "selected" in option.attrs:
                        self._formdata.add(
                            fieldname, option.attrs.get("value", option.text)
                        )
            else:
                if options:
                    self._formdata[fieldname] = options[0].attrs.get(
                        "value", options[0].text
                    )
                    for option in options:
                        if "selected" in option.attrs:
                            self._formdata[fieldname] = option.attrs.get(
                                "value", option.text
                            )
                            break

        # field textearea...

    def set(self, fieldname: str, value: str) -> Any:
        """
        Set a value to an input field.

        It works for checkbox and radio as well.
        Checkbox may contains many values.
        Options of select can't be set with this method, the select method must
        be used instead.
        """
        if fieldname not in self._formfields:
            raise ValueError(f'"{fieldname}" does not exists')
        if self._formfields[fieldname].node_name == "select":
            raise ValueError(f'"{fieldname}" is a <select>, use select() instead')

        if self._formfields[fieldname].attrs.get("type") == "checkbox":
            self._formdata.add(fieldname, value)
            return

        if self._formfields[fieldname].attrs.get("type") == "radio":
            radio = self._form.by_node_name(
                "input", attrs={"type": "radio", "value": value}
            )
            if not radio:
                raise ValueError(
                    f'radio "{fieldname}" does not contains {value} option'
                )

        self._formdata[fieldname] = value

    def unset(self, fieldname: str, value: str) -> Any:
        """Unset an element. Only works with checkbox."""
        if fieldname not in self._formfields:
            raise ValueError(f'"{fieldname}" does not exists')
        if self._formfields[fieldname].node_name != "input":
            raise ValueError(f'"{fieldname}" is not a checkbox')
        if self._formfields[fieldname].attrs.get("type") != "checkbox":
            raise ValueError(f'"{fieldname}" is not a checkbox')
        values = self._formdata.popall(fieldname)
        if value not in values:
            raise ValueError(f'"{value}" not in "{fieldname}"')
        for val in values:
            if val != value:
                self._formdata[fieldname] = val

    def select(self, fieldname: str, value: str) -> Any:
        """
        Select an option, if multiple, value is added, otherwise, value is replaced.
        """
        if fieldname not in self._formfields:
            raise ValueError(f'"{fieldname}" does not exists')
        field = self._formfields[fieldname]
        if field.node_name != "select":
            raise ValueError(f"{fieldname} is a {field!r}, " "use set() instead")

        for option in field.by_node_name("option"):
            if option.text == value.strip():
                if "multiple" in field.attrs:
                    self._formdata.add(fieldname, value)
                else:
                    self._formdata[fieldname] = option.attrs.get("value", option.text)
                break
        else:
            raise ValueError(f'No option {value} in <select name="{fieldname}">')

    def unselect(self, fieldname: str, value: str) -> Any:
        """
        Unselect an option if multiple, otherwise an exception is raised.
        """
        if fieldname not in self._formfields:
            raise ValueError(f'"{fieldname}" does not exists')
        field = self._formfields[fieldname]

        if field.node_name != "select":
            raise ValueError(
                f"{fieldname} is a {self._formfields[fieldname]!r}, "
                "use unset() for checkbox instead"
            )
        if "multiple" not in field.attrs:
            raise ValueError("only <select multiple> support unselect")

        for option in self._formfields[fieldname].by_node_name("option"):
            if option.text == value.strip():
                values = self._formdata.popall(fieldname)
                if value not in values:
                    raise ValueError(f'"{value}" not selected in "{fieldname}"')
                for val in values:
                    if val != value:
                        self._formdata[fieldname] = val
                break
        else:
            raise ValueError(f'No option {value} in <select name="{fieldname}">')

    def button(self, text: str, position: int = 0) -> "WebForm":
        """
        Simmulate a click on a button using the text of the button,

        and eventually a position. The button return the form and the submit()
        should be called directly.

        This is used in order to inject the value of the button in the form, usually
        done while many actions are available on a form.

        ::

            form.button("Go").submit()

        """
        buttons = self._form.get_all_by_text(text, node_name="button")
        if position >= len(buttons):
            pos = ""
            if position > 0:
                pos = f" at position {position}"
            raise ValueError(f'Button "{text}" not found{pos}')
        button = buttons[position]
        if "name" in button.attrs:
            self._formdata[button.attrs["name"]] = button.attrs.get("value", "")
        return self

    def submit(self, follow_redirects: bool = True) -> "WebResponse":
        """
        Submit the form as it has been previously filled out.
        """
        headers: dict[str, str] = {}
        target = (
            self._form.attrs.get("hx-post")
            or self._form.attrs.get("post")
            or self._origin
        )
        if "hx-post" in self._form.attrs:
            if hx_target := self._form.hx_target:
                headers["HX-Target"] = hx_target

        return self._client.post(
            target,
            data=self._formdata,
            headers=headers,
            follow_redirects=follow_redirects,
        )

    def __contains__(self, key: str) -> bool:
        """Test if a field exists in the form."""
        return key in self._formdata

    def __repr__(self) -> str:
        return repr(self._formdata)



"""
Widget for field of type bool.
"""

from .base import Widget


class BooleanWidget(Widget[bool]):
    """
    Widget for field of type bool.
    """

    template = """
    <pydantic_form.Widget :widget_id="id" :removable="removable">
      <div class="pt-4">
        <div class="flex items-center">
         <Checkbox :name="name" :id="id" :checked="value" value="1" />
         <Label :for="id" class="ms-2 text-base text-neutral-900 dark:text-white">
            {{title|safe}}
          </Label>
        </div>
       <pydantic_form.Error :text="error" />
      </div>
    </pydantic_form.Widget>
    """



from collections.abc import Sequence
from typing import Any

from markupsafe import Markup
from pydantic import Field

from fastlife.services.templates import AbstractTemplateRenderer

from .base import TWidget, TypeWrapper, Widget


class SequenceWidget(Widget[Sequence[TWidget]]):
    template = """
    <pydantic_form.Widget :widget_id="id" :removable="removable">
      <Details :id="id">
        <Summary :id="id + '-summary'">
          <H3 :class="H3_SUMMARY_CLASS">{{title}}</H3>
          <pydantic_form.Error :text="error" />
        </Summary>
        <div>
          {% set fnGetName = "get" + id.replace("-", "_") %}
          <script>
            function {{ fnGetName }} () {
              const el = document.getElementById("{{id}}-content");
              const len = el.dataset.length;
              el.dataset.length = parseInt(len) + 1;
              return "{{wrapped_type.name}}." + len;
            }
          </script>

          <div id="{{id}}-content" class="m-4"
            data-length="{{children_widgets|length|string}}">
            {% set container_id = id + "-children-container" %}
            <div id="{{container_id}}">
              {% for child in children_widgets %}
              {{ child }}
              {% endfor%}
            </div>
          </div>

          <div>
            {% set container_id = "#" + id + "-children-container" %}
            {% set add_id = id + "-add" %}
            {% set vals = 'js:{"name": '
                + fnGetName
                + '(), "token": "'
                + wrapped_type.token + '", "removable": true}' %}
            <Button type="button"
                :hx-target="container_id" hx-swap="beforeend"
                :id="add_id"
                :hx-vals="vals"
                :hx-get="wrapped_type.url">
              Add
            </Button>
          </div>
        </div>
      </Details>
    </pydantic_form.Widget>
    """

    item_type: type[Any]
    wrapped_type: TypeWrapper | None = Field(default=None)
    children_widgets: list[str] = Field(default=None)

    def build_item_type(self, route_prefix: str) -> TypeWrapper:
        return TypeWrapper(self.item_type, route_prefix, self.name, self.token)

    def to_html(self, renderer: "AbstractTemplateRenderer") -> Markup:
        """Return the html version."""
        self.wrapped_type = self.build_item_type(renderer.route_prefix)
        self.children_widgets = [
            Markup(item.to_html(renderer)) for item in self.value or []
        ]
        return Markup(renderer.render_template(self))



"""Resolution of python objects for dependency injection and more."""

import importlib.util
import inspect
from pathlib import Path
from types import ModuleType, UnionType
from typing import Any, Union


def resolve(value: str) -> Any:
    """return the attr from the syntax: package.module:attr."""
    module_name, attr_name = value.split(":", 2)
    spec = importlib.util.find_spec(module_name)
    if spec is None:
        raise ValueError(f"Module {module_name} not found")
    module = importlib.util.module_from_spec(spec)
    if spec.loader is None:
        # I can't figure out when the spec.loader is None,
        # just relying on typing here
        raise ValueError("No loader on spec")  # coverage: ignore
    spec.loader.exec_module(module)

    try:
        attr = getattr(module, attr_name)
    except AttributeError as exc:
        raise ValueError(
            f"Attribute {attr_name} not found in module {module_name}"
        ) from exc

    return attr


def resolve_extended(value: str) -> UnionType:
    """Resolve many types separed by a pipe (``|``), the union separator."""
    values = value.split("|")
    if len(values) == 1:
        return resolve(value)
    types = [resolve(t) for t in values if t != "builtins:NoneType"]
    return Union[tuple(types)]  # type: ignore # noqa: UP007


def resolve_path(value: str) -> str:
    """
    Resole a path on the disk from a python package name.

    This helper is used to find static assets inside a python package.
    """
    package_name, resource_name = value.split(":", 1)
    spec = importlib.util.find_spec(package_name)
    if not spec or not spec.origin:
        raise ValueError(f"{value} not found")
    package_path = spec.origin
    full_path = Path(package_path).parent / resource_name
    return str(full_path)


def resolve_package(mod: ModuleType) -> ModuleType:
    """
    Return the
    [regular package](https://docs.python.org/3/glossary.html#term-regular-package)
    of a module or itself if it is the ini file of a package.

    """

    # Compiled package has no __file__ attribute, ModuleType declare it as NoneType
    if not hasattr(mod, "__file__") or mod.__file__ is None:
        return mod

    module_path = Path(mod.__file__)
    if module_path.name == "__init__.py":
        return mod

    parent_module_name = mod.__name__.rsplit(".", 1)[0]
    parent_module = importlib.import_module(parent_module_name)
    return parent_module


def _strip_left_dots(s: str) -> tuple[str, int]:
    stripped_string = s.lstrip(".")
    num_stripped_dots = len(s) - len(stripped_string)
    return stripped_string, num_stripped_dots - 1


def _get_parent(pkg: str, num_parents: int) -> str:
    if num_parents == 0:
        return pkg
    segments = pkg.split(".")
    return ".".join(segments[:-num_parents])


def resolve_maybe_relative(mod: str, stack_depth: int = 1) -> ModuleType:
    """
    Resolve a module, maybe relative to the stack frame and import it.

    :param mod: the module to import. starts with a dot if it is relative.
    :param stack_depth: relative to which module in the stack.
        used to do an api that call it instead of resolve the module directly.
    :return: the imported module
    """
    if mod.startswith("."):
        caller_module = inspect.getmodule(inspect.stack()[stack_depth][0])
        # we could do an assert here but caller_module could really be none ?
        parent_module = resolve_package(caller_module)  # type: ignore
        package = parent_module.__name__
        mod, count = _strip_left_dots(mod)
        package = _get_parent(package, count)
        mod = f"{package}.{mod}".rstrip(".")

    module = importlib.import_module(mod)
    return module



from typing import Annotated

from fastapi import Depends

from fastlife.request.request import Request
from fastlife.services.translations import Localizer as RequestLocalizer


def get_localizer(request: Request) -> RequestLocalizer:
    return request.registry.localizer(request)


Localizer = Annotated[RequestLocalizer, Depends(get_localizer)]



"""Type inference."""

from types import UnionType
from typing import Any, Union, get_origin

from pydantic import BaseModel


def is_complex_type(typ: type[Any]) -> bool:
    """
    Used to detect complex type such as Mapping, Sequence and pydantic BaseModel.

    This method cannot be used outside pydantic serialization.
    """
    return bool(get_origin(typ) or issubclass(typ, BaseModel))


def is_union(typ: type[Any]) -> bool:
    """Used to detect unions like Optional[T], Union[T, U] or T | U."""
    type_origin = get_origin(typ)
    if type_origin:
        if type_origin is Union:  # Optional[T]
            return True

        if type_origin is UnionType:  # T | U
            return True
    return False



"""Security policy."""

import abc
import logging
from collections.abc import Callable, Coroutine
from typing import Annotated, Any, Generic, Literal, TypeVar
from uuid import UUID

from fastapi import Depends, HTTPException
from starlette.status import HTTP_401_UNAUTHORIZED, HTTP_403_FORBIDDEN

from fastlife import GenericRequest, get_request
from fastlife.config.registry import TRegistry

CheckPermissionHook = Callable[..., Coroutine[Any, Any, None]] | Callable[..., None]
CheckPermission = Callable[[str], CheckPermissionHook]

TUser = TypeVar("TUser")

log = logging.getLogger(__name__)


class Unauthorized(HTTPException):
    """An exception raised to stop a request exectution and return an HTTP Error."""

    def __init__(
        self,
        status_code: int = HTTP_401_UNAUTHORIZED,
        detail: str = "Unauthorized",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code, detail, headers)


class Forbidden(HTTPException):
    """An exception raised to stop a request exectution and return an HTTP Error."""

    def __init__(
        self,
        status_code: int = HTTP_403_FORBIDDEN,
        detail: str = "Forbidden",
        headers: dict[str, str] | None = None,
    ) -> None:
        super().__init__(status_code, detail, headers)


class BoolMeta(type):
    def __bool__(cls) -> bool:
        return cls.kind == "allowed"  # type: ignore

    def __repr__(cls) -> str:
        return cls.reason  # type: ignore


class HasPermission(int, metaclass=BoolMeta):
    """
    A type used to know if a permission is allowed or not.

    It behave has a boolean, but 3 possibilities exists defind has 3 sub-types
    {class}`Allowed` {class}`Unauthenticated` or {class}`Denied`.

    In many cases Unauthenticated call may redirect to a login page,
    where authenticated user are not redirected. they have an error message,
    or the frontend may use the information to adapt its interface.
    """

    kind: Literal["allowed", "unauthenticated", "denied"]
    reason: str

    def __new__(cls, reason: str) -> "HasPermission":
        instance = super().__new__(cls)
        instance.reason = reason
        return instance

    def __repr__(self) -> str:
        return self.reason

    def __bool__(self) -> bool:
        return self.kind == "allowed"


class Allowed(HasPermission):
    """Represent a permission check result that is allowed."""

    kind = "allowed"
    reason = "Allowed"


class Unauthenticated(HasPermission):
    """
    Represent a permission check result that is not allowed due to
    missing authentication mechanism.
    """

    kind = "unauthenticated"
    reason = "Authentication required"


class Denied(HasPermission):
    """
    Represent a permission check result that is not allowed due to lack of permission.
    """

    kind = "denied"
    reason = "Access denied to this resource"


class AbstractSecurityPolicy(abc.ABC, Generic[TUser, TRegistry]):
    """Security policy base classe."""

    Forbidden = Forbidden
    """The exception raised if the user identified is not granted."""
    Unauthorized = Unauthorized
    """The exception raised if no user has been identified."""

    request: GenericRequest[TRegistry]
    """Request where the security policy is applied."""

    def __init__(
        self, request: Annotated[GenericRequest[TRegistry], Depends(get_request)]
    ):
        """
        Build the security policy.

        When implementing a security policy, multiple parameters can be added
        to the constructor as FastAPI dependencies, using the `Depends` FastAPI
        annotation.
        The security policy is installed has a depenency of the router that hold
        a route prefix of the application.
        """
        self.request = request
        self.request.security_policy = self  # we do backref to implement has_permission

    @abc.abstractmethod
    async def identity(self) -> TUser | None:
        """
        Return app-specific user object or raise an HTTPException.
        """

    @abc.abstractmethod
    async def authenticated_userid(self) -> str | UUID | None:
        """
        Return app-specific user object or raise an HTTPException.
        """

    @abc.abstractmethod
    async def has_permission(
        self, permission: str
    ) -> HasPermission | type[HasPermission]:
        """Allow access to everything if signed in."""

    @abc.abstractmethod
    async def remember(self, user: TUser) -> None:
        """Save the user identity in the request session."""

    @abc.abstractmethod
    async def forget(self) -> None:
        """Destroy the request session."""


class InsecurePolicy(AbstractSecurityPolicy[None, Any]):
    """
    An implementation of the security policy made for explicit unsecured access.

    Setting a permission on a view require a security policy, if not set, accessing
    to a view will raise a RuntimeError. To bypass this error for testing purpose
    or your own reason, the InsecurePolicy has to be set to the configurator.
    """

    async def identity(self) -> None:
        """Nobodies is identified."""
        return None

    async def authenticated_userid(self) -> str | UUID:
        """An uuid mades of 0."""
        return UUID(int=0)

    async def has_permission(
        self, permission: str
    ) -> HasPermission | type[HasPermission]:
        """Access is allways granted."""
        return Allowed

    async def remember(self, user: None) -> None:
        """Do nothing."""

    async def forget(self) -> None:
        """Do nothing."""



from collections.abc import Mapping
from typing import TYPE_CHECKING, Generic, TypeVar

from fastlife.services.locale_negociator import LocaleNegociator, default_negociator
from fastlife.services.translations import LocalizerFactory

if TYPE_CHECKING:
    from fastlife.services.templates import (  # coverage: ignore
        AbstractTemplateRendererFactory,  # coverage: ignore
    )  # coverage: ignore

from .settings import Settings

TSettings = TypeVar("TSettings", bound=Settings, covariant=True)
"""
A TypeVar used to override the DefaultRegistry to add more helpers in the registry.
"""


class GenericRegistry(Generic[TSettings]):
    """
    Application registry for fastlife dependency injection.
    It is initialized by the configurator and accessed by the `fastlife.Registry`.
    """

    settings: Settings
    renderers: Mapping[str, "AbstractTemplateRendererFactory"]
    locale_negociator: LocaleNegociator
    localizer: LocalizerFactory

    def __init__(self, settings: Settings) -> None:
        self.settings = settings
        self.locale_negociator = default_negociator(self.settings)
        self.renderers = {}
        self.localizer = LocalizerFactory()

    def get_renderer(self, template: str) -> "AbstractTemplateRendererFactory":
        for key, val in self.renderers.items():
            if template.endswith(key):
                return val
        raise RuntimeError(f"No renderer registered for template {template}")


DefaultRegistry = GenericRegistry[Settings]
"""
The default registry until you need to inject more component in the registry.
"""


TRegistry = TypeVar("TRegistry", bound=DefaultRegistry, covariant=True)
"""
A TypeVar used to override the DefaultRegistry to add more helpers in the registry.
"""



from collections.abc import Sequence

from pydantic import Field

from fastlife.domain.model.types import Builtins

from .base import Widget


class TextWidget(Widget[Builtins]):
    """
    Widget for text like field (email, ...).
    """

    template = """
    <pydantic_form.Widget :widget_id="id" :removable="removable">
      <div class="pt-4">
        <Label :for="id">{{title}}</Label>
        <pydantic_form.Error :text="error" />
        <Input :name="name" :value="value" :type="input_type" :id="id"
          :aria-label="aria_label" :placeholder="placeholder" />
        <pydantic_form.Hint :text="hint" />
      </div>
    </pydantic_form.Widget>
    """

    input_type: str = Field(default="text")
    placeholder: str | None = Field(default=None)


class TextareaWidget(Widget[Sequence[str]]):
    """
    Render a Textearea for a string or event a sequence of string.

    ```
    from fastlife.adapters.jinjax.widgets.base import CustomWidget
    from fastlife.adapters.jinjax.widgets.text import TextareaWidget
    from pydantic import BaseModel, Field, field_validator

    class TaggedParagraphForm(BaseModel):
        paragraph: Annotated[str, CustomWidget(TextareaWidget)] = Field(...)
        tags: Annotated[Sequence[str], CustomWidget(TextareaWidget)] = Field(
            default_factory=list,
            title="Tags",
            description="One tag per line",
        )

        @field_validator("tags", mode="before")
        def split(cls, s: Any) -> Sequence[str]:
            return s.split() if s else []
    ```
    """

    template = """
    <pydantic_form.Widget :widget_id="id" :removable="removable">
      <div class="pt-4">
        <Label :for="id">{{title}}</Label>
        <pydantic_form.Error :text="error" />
        <Textarea :name="name" :id="id" :aria-label="aria_label">
            {%- if v is string -%}
            {{- v -}}}
            {%- else -%}
            {%- for v in value %}{{v}}
    {% endfor -%}
            {% endif %}
        </Textarea>
        <pydantic_form.Hint :text="hint" />
      </div>
    </pydantic_form.Widget>
    """

    placeholder: str = Field(default="")



"""
Widget for field of type Union.
"""

from collections.abc import Sequence
from typing import Union

from markupsafe import Markup
from pydantic import BaseModel, Field

from fastlife.services.templates import AbstractTemplateRenderer

from .base import TWidget, TypeWrapper, Widget


class UnionWidget(Widget[TWidget]):
    """
    Widget for union types.
    """

    template = """
    <pydantic_form.Widget :widget_id="id" :removable="removable">
      <div id="{{id}}">
        <Details>
          <Summary :id="id + '-union-summary'">
            <H3 :class="H3_SUMMARY_CLASS">{{title}}</H3>
            <pydantic_form.Error :text="error" />
          </Summary>
          <div hx-sync="this" id="{{id}}-child">
            {% if child %}
            {{ child }}
            {% else %}
            {% for typ in types %}
            <Button type="button"
                hx-target="closest div"
                :hx-get="typ.url"
                :hx-vals="typ.params|tojson"
                :id="typ.id"
                onclick={{ "document.getElementById('" + id + "-remove-btn').hidden=false" }}
              :class="SECONDARY_BUTTON_CLASS">{{typ.title}}</Button>
            {% endfor %}
            {% endif %}
          </div>
          <Button type="button" :id="id + '-remove-btn'" :hx-target="'#' + id"
            :hx-vals="parent_type.params|tojson" :hx-get="parent_type.url" :hidden="not child"
            :class="SECONDARY_BUTTON_CLASS">
            Remove
          </Button>
        </Details>
      </div>
    </pydantic_form.Widget>
    """

    children_types: Sequence[type[BaseModel]]
    parent_type: TypeWrapper | None = Field(default=None)

    types: Sequence[TypeWrapper] | None = Field(default=None)
    child: str = Field(default="")

    def build_types(self, route_prefix: str) -> Sequence[TypeWrapper]:
        """Wrap types in the union in order to get the in their own widgets."""
        return [
            TypeWrapper(typ, route_prefix, self.name, self.token)
            for typ in self.children_types
        ]

    def to_html(self, renderer: "AbstractTemplateRenderer") -> Markup:
        """Return the html version."""
        self.child = Markup(self.value.to_html(renderer)) if self.value else ""
        self.types = self.build_types(renderer.route_prefix)
        self.parent_type = TypeWrapper(
            Union[tuple(self.children_types)],  # type: ignore # noqa: UP007
            renderer.route_prefix,
            self.name,
            self.token,
            title=self.title,
        )
        return Markup(renderer.render_template(self))



"""Find the localization gor the given request."""

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from fastlife.config.settings import Settings

LocaleName = str
"""The LocaleName is a locale such as en, fr that will be consume for translations."""

if TYPE_CHECKING:
    from fastlife.request.request import GenericRequest  # coverage: ignore

    LocaleNegociator = Callable[[GenericRequest[Any]], LocaleName]  # coverage: ignore
    """Interface to implement to negociate a locale"""  # coverage: ignore
else:
    LocaleNegociator = Any


def default_negociator(settings: Settings) -> LocaleNegociator:
    """The default local negociator return the locale set in the conf."""

    def locale_negociator(request: "GenericRequest[Any]") -> str:
        return settings.default_locale

    return locale_negociator
